type Fn = {
    name: string;
    fn: (...args: any) => any;
};
type CallContext = {
    that: unknown;
    args: IArguments;
};
type ExecuteState = {
    state: "execute";
};
type ErrorState = {
    state: "error";
    error: unknown;
};
type NormalState = {
    state: "result";
    result: unknown;
};
type InstrumentHandler = {
    name?: string;
    onExecute?: (context: Fn & CallContext & ExecuteState) => void | Partial<CallContext & Fn & ExecuteState>;
    onError?: (context: Fn & CallContext & ErrorState & {
        isPromise: boolean;
    }) => void | ErrorState | NormalState;
    onResult?: (context: Fn & CallContext & NormalState & {
        isPromise: boolean;
    }) => void | ErrorState | NormalState;
};
type CreateInstrumentHandler = InstrumentHandler | (() => InstrumentHandler);
declare const createInstrument: (opts: CreateInstrumentHandler) => InstrumentHandler;
type InstrumentFunction = <X extends Fn["fn"]>(fn: X, name?: string) => X;
declare function composeInstrument(...ins: CreateInstrumentHandler[]): InstrumentFunction | undefined;

type Provider<Provide, Input = unknown> = (input: Input) => Provide | Promise<Provide>;
type ProviderOption<Provide> = {
    name?: string;
    instrument?: InstrumentFunction;
    mode?: 'prototype' | 'singleton';
    onExecute?: <V>(provide: Provide, execution: Provider<V, Provide>) => V | Promise<V>;
    onError?: (error: unknown, value?: Provide) => (void | Promise<void>);
};
declare const defaultProviderOptions: ProviderOption<any>;
declare function setInstrument(inst: CreateInstrumentHandler): void;
type Executor<Value, Dependent> = {
    set(value: Value): void;
    get(deps?: Executor<Dependent, any>): Promise<Value>;
    execute<Output>(execution: Provider<Output, Value>): Promise<Output>;
    readonly [x: symbol]: true;
};
declare function make<Provide>(provider: Provider<Provide>, options?: ProviderOption<Provide>): Executor<Provide, void>;
declare function make<Dependency, Provide>(dependencies: Executor<Dependency, unknown>, provider: Provider<Provide, Dependency>, options?: ProviderOption<Provide>): Executor<Provide, Dependency>;
declare function create<Provide>(provider: Provider<Provide>, options?: ProviderOption<Provide>): Executor<Provide, void>;
declare function create<Provide, Dependent>(provider: Provider<Provide, Dependent>, dependent?: Executor<Dependent, unknown>, options?: ProviderOption<Provide>): Executor<Provide, Dependent>;
declare const value: <Provide>(value: Provide, options?: Omit<ProviderOption<Provide>, "mode"> | undefined) => Executor<Provide, void>;
declare const template: <Dependent>(dependent: Executor<Dependent, unknown>) => <Fn extends (...input: any[]) => any, Input extends any[] = Parameters<Fn>>(factory: (dependent: Dependent, ...params: Input) => ReturnType<Fn>) => (...params: Input) => Promise<ReturnType<Fn>>;
declare const prepare: <Dependent, Input extends any[], Output>(provider: (provide: Dependent, ...input: Input) => Output, dep: Executor<Dependent, unknown>, options?: ProviderOption<Output> | undefined) => (...input: Input) => Promise<Awaited<Output>>;
declare function execute<Output>(executable: Provider<Output>, options?: ProviderOption<Output>): Promise<Awaited<Output>>;
declare function execute<Output, Dependent>(executable: Provider<Output, Dependent>, dependent: Executor<Dependent, unknown>, options?: ProviderOption<Output>): Promise<Awaited<Output>>;
type inferProvide<T> = T extends Executor<infer S, any> ? S : never;
declare const combine: <L extends Record<string, Executor<any, any>>>(layout: L) => Executor<{ [key in keyof L]: inferProvide<L[key]>; }, unknown>;

export { CreateInstrumentHandler, Executor, InstrumentFunction, InstrumentHandler, ProviderOption, combine, composeInstrument, create, createInstrument, defaultProviderOptions, execute, inferProvide, make, prepare, setInstrument, template, value };
