var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);

// src/index.ts
var src_exports = {};
__export(src_exports, {
  combine: () => combine,
  composeInstrument: () => composeInstrument,
  create: () => create,
  createInstrument: () => createInstrument,
  defaultProviderOptions: () => defaultProviderOptions,
  execute: () => execute,
  make: () => make,
  prepare: () => prepare,
  setInstrument: () => setInstrument,
  template: () => template,
  value: () => value
});
module.exports = __toCommonJS(src_exports);

// src/instrument.ts
function isPromise(value2) {
  return value2 === Promise.resolve(value2);
}
var createInstrument = (opts) => {
  return typeof opts === "function" ? opts() : opts;
};
function createInstrumentor(opts) {
  return function(fn, fnName) {
    let name = opts?.name || fn.name || fnName || "anonymous";
    const handler = typeof opts === "function" ? opts() : opts;
    const f = function() {
      let that = this;
      let args = arguments;
      let func = fn;
      try {
        const executeState = handler.onExecute?.({ name, fn, that, args, state: "execute" });
        if (executeState !== void 0) {
          that = executeState.that || that;
          args = executeState.args || args;
          func = executeState.fn || func;
          name = executeState.name || name;
        }
        let result = func.apply(that, args);
        if (isPromise(result)) {
          return result.then((next) => {
            const promiseResultState = handler.onResult?.({
              name,
              fn: func,
              that,
              args,
              state: "result",
              result: next,
              isPromise: true
            });
            if (promiseResultState?.state === "result") {
              return promiseResultState.result;
            }
            if (promiseResultState?.state === "error") {
              throw promiseResultState.error;
            }
            return next;
          }).catch((error) => {
            const errorState = handler.onError?.({
              name,
              fn: func,
              that,
              args,
              error,
              state: "error",
              isPromise: true
            });
            if (errorState?.state === "error") {
              throw errorState.error;
            }
            if (errorState?.state === "result") {
              return errorState.result;
            }
            throw error;
          });
        } else {
          const resultState = handler.onResult?.({
            name,
            fn: func,
            that,
            args,
            state: "result",
            result: void 0,
            isPromise: false
          });
          if (resultState?.state === "error") {
            throw resultState.error;
          }
          if (resultState?.state === "result") {
            return resultState.result;
          }
          return result;
        }
      } catch (error) {
        const errorState = handler.onError?.({ name, fn: func, that, args, error, state: "error", isPromise: false });
        if (errorState?.state === "error") {
          throw errorState.error;
        }
        if (errorState?.state === "result") {
          return errorState.result;
        }
        throw error;
      }
    };
    Object.defineProperty(f, "name", {
      value: name,
      writable: false
    });
    return f;
  };
}
function instrument(container, instrumentFn, maxLevel = 1, currentLevel = 0) {
  if (container === null || container === void 0)
    return container;
  if (currentLevel > maxLevel)
    return container;
  if (
    // there can be more
    typeof container !== "object" || Array.isArray(container)
  ) {
    return container;
  }
  Object.keys(container).forEach((key) => {
    const instance = container[key];
    if (typeof instance === "function") {
      container[key] = instrumentFn(instance, key);
    } else {
      container[key] = instrument(container[key], instrumentFn, maxLevel, currentLevel + 1);
    }
  });
  return container;
}
function nextInstrument(prev, inst) {
  const next = createInstrumentor(createInstrument(inst));
  const mixin = (fn, name) => {
    const instrumented = prev?.(fn, name) || fn;
    return next(instrumented, name);
  };
  return mixin;
}
function composeInstrument(...ins) {
  return ins.reduceRight((prev, next) => {
    return nextInstrument(prev, next);
  }, void 0);
}

// src/index.ts
var defaultProviderOptions = {
  mode: "singleton",
  instrument: createInstrumentor({})
};
function setInstrument(inst) {
  defaultProviderOptions.instrument = nextInstrument(defaultProviderOptions.instrument, inst);
}
var executorSymbol = Symbol.for("$submodule");
function isExecutor(obj) {
  return obj?.[executorSymbol];
}
function extract(executable, secondParam, thirdParam) {
  return isExecutor(secondParam) ? { provider: executable, dependent: secondParam, options: thirdParam } : { provider: executable, dependent: void 0, options: secondParam || thirdParam };
}
function forceObjectProp(target, name, value2) {
  Object.defineProperty(target, name, { value: value2, writable: false });
}
function make(firstParam, secondParam, thirdParam) {
  if (isExecutor(firstParam)) {
    if (secondParam === void 0) {
      throw new Error("invalid state, provider is required");
    }
    if (typeof secondParam !== "function") {
      throw new Error("invalid state, provider is required");
    }
    return create(secondParam, firstParam, thirdParam);
  } else {
    if (typeof secondParam === "function") {
      throw new Error("invalid state, second param must be option or undefined");
    }
    return create(firstParam, void 0, secondParam);
  }
}
function create(providerParam, secondParam, thirdParam) {
  const { provider, dependent, options } = extract(providerParam, secondParam, thirdParam);
  const opts = { ...defaultProviderOptions, ...options };
  let depExecutor = dependent;
  let loader = void 0;
  const name = opts?.name || provider.name || "anonymous";
  async function load(inject) {
    try {
      const actualized = inject ? await inject.get() : await depExecutor?.get();
      if (provider.length > 0 && actualized === void 0) {
        throw new Error(`invalid state, provider ${provider.toString()} requires dependent but not provided`);
      }
      return await provider(actualized);
    } catch (e) {
      await opts.onError?.(e);
      throw e;
    }
  }
  async function get(dep) {
    if (opts?.mode === "singleton") {
      if (loader === void 0) {
        loader = load(dep);
      }
      return await loader;
    }
    return await load(dep);
  }
  async function execute2(execution) {
    let value2 = void 0;
    try {
      const value3 = await get();
      return opts.onExecute ? await opts.onExecute(value3, execution) : execution(value3);
    } catch (e) {
      await opts.onError?.(e, value2);
      throw e;
    }
  }
  function set(value2) {
    if (loader === void 0) {
      loader = Promise.resolve(value2);
    }
  }
  const executor = { get, set, execute: execute2, [Symbol.for("$submodule")]: true };
  forceObjectProp(executor.get, "name", `${name}.get`);
  forceObjectProp(executor.set, "name", `${name}.set`);
  forceObjectProp(executor.execute, "name", `${name}.execute`);
  if (opts.instrument) {
    instrument(executor, opts.instrument);
  }
  return executor;
}
var value = (value2, options) => create(() => value2, options);
var template = (dependent) => (factory) => {
  return (...params) => dependent.execute((dep) => factory(dep, ...params));
};
var prepare = (provider, dep, options) => {
  return async function() {
    const that = this;
    const args = arguments;
    return execute(async () => provider.call(that, await dep.get(), ...args), options);
  };
};
async function execute(executable, secondParam, thirdParam) {
  const { provider, dependent, options } = extract(executable, secondParam, thirdParam);
  return await create(() => dependent ? dependent.execute(provider) : provider(), options).get();
}
var combine = function(layout) {
  return create(async () => {
    const layoutPromise = Object.entries(layout).map(([key, executor]) => executor.get().then((value2) => [key, value2]));
    const result = Object.fromEntries(await Promise.all(layoutPromise));
    return result;
  });
};
// Annotate the CommonJS export names for ESM import in node:
0 && (module.exports = {
  combine,
  composeInstrument,
  create,
  createInstrument,
  defaultProviderOptions,
  execute,
  make,
  prepare,
  setInstrument,
  template,
  value
});
//# sourceMappingURL=index.js.map